from ._openai_api import OpenAIAPIProvider
import os


class Vercel(OpenAIAPIProvider):
    def __init__(self, model: str, api_key: str | None = None):
        api_key = (
            api_key
            or os.environ.get("VERCEL_AI_GATEWAY_API_KEY")
            or os.environ.get("VERCEL_API_KEY")
        )
        if not api_key:
            raise ValueError(
                "VERCEL_AI_GATEWAY_API_KEY or VERCEL_API_KEY environment variable not set"
            )
        base_url = "https://ai-gateway.vercel.sh/v1"
        super().__init__(name="vercel", model=model, api_key=api_key, base_url=base_url)
        self.extra_body["modalities"] = ["text", "image"]



from typing import Any, override
from ._openai_api import OpenAIAPIProvider
import os


class Ollama(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        base_url = base_url or os.environ.get(
            "OLLAMA_BASE_URL",
            os.environ.get("OPENAI_BASE_URL", "http://localhost:11434/v1"),
        )
        api_key = api_key or os.environ.get("OLLAMA_API_KEY", "dummy")
        super().__init__(name="ollama", model=model, api_key=api_key, base_url=base_url)

    @override
    def get_reasoning_args(
        self,
        enabled: bool | None,
        effort: str | None,
        exclude: bool | None,
        max_tokens: int | None,
    ) -> dict[str, Any]:
        if not enabled:
            return {}
        return {
            "think": enabled,
        }



from ..utils.decorators import tool
from agentia.tools import Plugin, PluginInitError


__all__ = [
    "Plugin",
    "PluginInitError",
    "tool",
]

# Plugins
from .calc import Calculator
from .clock import Clock
from .code_runner import CodeRunner
from .skills import Skills
from .memory import Memory
from .search import Search
from .web import Web
from .bash import Bash



from typing import TYPE_CHECKING, AsyncGenerator
import httpx
from agentia.llm import LLMOptions
from agentia.llm.completion import ChatCompletion
from agentia.llm.stream import ChatCompletionEvents, ChatCompletionStream
from agentia.spec import *
from agentia.tools.tools import ToolSet

if TYPE_CHECKING:
    from agentia.agent import Agent


async def __process_tool_calls(
    agent: "Agent", tool_calls: list[ToolCall], tools: ToolSet, parallel: bool
) -> tuple[ToolMessage, list[ToolCallResponse]]:
    assert tools is not None, "No tools provided"
    tool_responses = await tools.run(agent, tool_calls, parallel)
    tool_msg_parts: list[MessagePartToolResult] = []
    for tr in tool_responses:
        mptr = MessagePartToolResult(
            tool_call_id=tr.tool_call_id,
            tool_name=tr.tool_name,
            input=tr.input,
            output=tr.output,
            output_files=tr.output_files,
        )
        tool_msg_parts.append(mptr)
    tool_msg = ToolMessage(content=tool_msg_parts)
    return tool_msg, tool_responses


def run_agent_loop(agent: "Agent", options: LLMOptions) -> ChatCompletion:
    tools = agent.tools

    async def gen() -> AsyncGenerator[AssistantMessage | ToolMessage, None]:
        await tools.init()
        messages = agent.history.get()
        async with httpx.AsyncClient() as client:
            while True:
                result = await agent.provider.generate(
                    messages=messages, tools=tools, options=options, client=client
                )
                c.usage += result.usage
                c._add_new_message(result.message)
                yield result.message
                # Add new messages
                messages.append(result.message)
                tool_calls = result.message.tool_calls
                if result.finish_reason != "tool-calls":
                    break
                # Call tools and continue
                tool_msg, _ = await __process_tool_calls(
                    agent,
                    tool_calls,
                    tools=tools,
                    parallel=options.parallel_tool_calls or False,
                )
                yield tool_msg
                messages.append(tool_msg)
                c._add_new_message(tool_msg)

        agent.history.add(*c.new_messages)

    async def gen_with_mcp_context() -> (
        AsyncGenerator[AssistantMessage | ToolMessage, None]
    ):
        from agentia.tools.mcp import MCPContext

        async with MCPContext() as _ctx:
            agent._temp_mcp_context = _ctx
            async for msg in gen():
                yield msg
            agent._temp_mcp_context = None

    c = ChatCompletion(gen_with_mcp_context(), agent)
    return c


def run_agent_loop_streamed(
    agent: "Agent", events: bool, options: LLMOptions
) -> ChatCompletionStream | ChatCompletionEvents:
    tools = agent.tools

    async def gen() -> AsyncGenerator[StreamPart, None]:
        await tools.init()
        messages = agent.history.get()
        last_finish_reason: FinishReason = "unknown"

        async with httpx.AsyncClient() as client:
            stream_started = False
            while True:
                tool_calls: list[ToolCall] = []
                parts: list[AssistantMessagePart] = []
                last_msg = ""
                last_reasoning = ""
                async for part in agent.provider.stream(
                    messages, tools, options, client
                ):

                    match part.type:
                        case "stream-start":
                            if not stream_started:
                                stream_started = True
                                yield part
                            yield StreamPartMessageStart(role="assistant")
                        case "text-start":
                            last_msg = ""
                        case "text-delta":
                            last_msg += part.delta
                        case "text-end":
                            parts.append(MessagePartText(text=last_msg))
                        case "reasoning-start":
                            last_reasoning = ""
                        case "reasoning-delta":
                            last_reasoning += part.delta
                        case "reasoning-end":
                            parts.append(MessagePartReasoning(text=last_reasoning))
                        case "tool-call":
                            parts.append(
                                MessagePartToolCall(
                                    tool_call_id=part.tool_call_id,
                                    tool_name=part.tool_name,
                                    input=part.input,
                                    provider_executed=part.provider_executed,
                                )
                            )
                            if not part.provider_executed:
                                tool_calls.append(part)

                    if part.type == "stream-end":
                        s.usage += part.usage
                        last_finish_reason = part.finish_reason
                    elif part.type != "stream-start":
                        yield part
                msg = AssistantMessage(content=parts)
                yield StreamPartMessageEnd(role="assistant", message=msg)
                messages.append(msg)
                s._add_new_message(msg)
                if not tool_calls:
                    break
                # Call tools and continue
                yield StreamPartMessageStart(role="tool")
                tool_msg, tool_responses = await __process_tool_calls(
                    agent,
                    tool_calls,
                    tools,
                    options.parallel_tool_calls or False,
                )
                for tr in tool_responses:
                    yield tr
                yield StreamPartMessageEnd(role="tool", message=tool_msg)
                messages.append(tool_msg)
                s._add_new_message(tool_msg)
        s.finish_reason = last_finish_reason
        yield StreamPartStreamEnd(usage=s.usage, finish_reason=last_finish_reason)

        agent.history.add(*s.new_messages)

    async def gen_with_mcp_context() -> AsyncGenerator[StreamPart, None]:
        from agentia.tools.mcp import MCPContext

        async with MCPContext() as _ctx:
            agent._temp_mcp_context = _ctx
            async for part in gen():
                yield part
            agent._temp_mcp_context = None

    if events:
        s = ChatCompletionEvents(gen_with_mcp_context(), agent)
    else:
        s = ChatCompletionStream(gen_with_mcp_context(), agent)
    return s



from ._openai_api import OpenAIAPIProvider
import os


class OpenAI(OpenAIAPIProvider):
    def __init__(
        self, model: str, api_key: str | None = None, base_url: str | None = None
    ):
        api_key = api_key or os.environ.get("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("OPENAI_API_KEY environment variable not set")
        base_url = base_url or os.environ.get(
            "OPENAI_BASE_URL", "https://api.openai.com/v1"
        )
        super().__init__(name="openai", model=model, api_key=api_key, base_url=base_url)



from . import Plugin
from pathlib import Path
from typing import Sequence, override
import frontmatter
from pydantic import BaseModel

INSTRUCTIONS_TEMPLATE = """\
You have access to a set of skills to extend your ability.
To use a skill, read its SKILL.md file carefully and follow the instructions.
Load only the necessary skills to complete the tasks.

# AVAILABLE SKILLS
{skills}
"""


class Skill(BaseModel):
    path: Path
    name: str
    description: str
    loaded: bool = False


class Skills(Plugin):
    def __init__(self, search_paths: Sequence[Path | str] | None = None) -> None:
        if not search_paths:
            # default search path
            search_paths = [
                Path.cwd() / ".skills",
                Path.cwd() / ".agentia" / "skills",
                Path.home() / ".config" / "agentia" / "skills",
            ]
        self.search_paths = [Path(p) for p in search_paths]

    def load_all_skills(self) -> dict[str, Skill]:
        skills: dict[str, Skill] = {}
        for path in self.search_paths:
            if not path.is_dir():
                continue
            for skill in self.discover_skills(path):
                skills[skill.name] = skill
        return skills

    def is_skill(self, path: Path) -> bool:
        if path.name.startswith(".") or path.name.startswith("_"):
            return False
        if not (path / "SKILL.md").is_file():
            return False
        return True

    def discover_skills(self, path: Path) -> list[Skill]:
        """Recursively discover skills from the given path."""
        assert path.is_dir(), f"'{path}' is not a directory"
        if self.is_skill(path):
            return [self.load_skill(path)]
        skills = []
        for child in path.iterdir():
            if child.is_dir():
                skills.extend(self.discover_skills(child))
        return skills

    def load_skill(self, path: Path) -> Skill:
        path = path.resolve()
        md = path / "SKILL.md"
        assert md.is_file(), f"Skill metadata file '{md}' does not exist"
        doc = frontmatter.loads(md.read_text(encoding="utf-8"))
        name = doc.metadata.get("name", path.stem)
        assert isinstance(name, str), f"Skill name must be a string in '{md}'"
        description = doc.metadata.get("description", "")
        assert isinstance(
            description, str
        ), f"Skill description must be a string in '{md}'"
        return Skill(path=path, name=name, description=description)

    @override
    def get_instructions(self) -> str:
        skills = self.load_all_skills()
        s = ""
        for skill in skills.values():
            s += f"- **{skill.name}**:\n    * path: {str(skill.path/"SKILL.md")}\n    * description: {skill.description}\n\n"
        return INSTRUCTIONS_TEMPLATE.format(skills=s)



from __future__ import annotations

import logging

LOGGER = logging.getLogger("agentia")


def init_logging(level: logging._Level = logging.INFO):
    logging.basicConfig(
        level=level, format="[%(asctime)s][%(levelname)-8s][%(name)s] %(message)s"
    )
    logging.getLogger("httpx").setLevel(logging.WARNING)


from dotenv import load_dotenv

from .agent import Agent
from .spec import *
from .utils.decorators import magic, ImageUrl
from .utils.decorators import tool
from .plugins import Plugin, PluginInitError
from .plugins.skills import Skills, Skill
from . import plugins
from .tools import ProviderTool, Tool, Tools, MCP, MCPContext, ToolResult
from .history import History
from .llm import LLMOptions, ReasoningOptions
from .llm.completion import ChatCompletion
from .llm.stream import (
    TextStream,
    ReasoningStream,
    MessageStream,
    ChatCompletionStream,
    ChatCompletionEvents,
)


__all__ = [
    # submodules
    "plugins",
    # agent
    "Agent",
    "History",
    # llm
    "ChatCompletion",
    "TextStream",
    "ReasoningStream",
    "MessageStream",
    "ChatCompletionStream",
    "ChatCompletionEvents",
    "LLMOptions",
    "ReasoningOptions",
    # spec
    "ProviderOptions",
    "ProviderMetadata",
    "DataContent",
    "FinishReason",
    "ToolChoice",
    "Usage",
    "ToolCall",
    "ToolCallResponse",
    "File",
    "Annotation",
    "UserConsentRequest",
    "MessageBase",
    "SystemMessage",
    "MessagePartBase",
    "MessagePartText",
    "MessagePartReasoning",
    "MessagePartFile",
    "MessagePartToolCall",
    "MessagePartToolResult",
    "MessagePart",
    "UserMessagePart",
    "AssistantMessagePart",
    "UserMessage",
    "ObjectType",
    "AssistantMessage",
    "ToolMessage",
    "Message",
    "NonSystemMessage",
    "ResponseFormatText",
    "ResponseFormatJson",
    "ResponseFormat",
    "StreamPartTextStart",
    "StreamPartTextDelta",
    "StreamPartTextEnd",
    "StreamPartReasoningStart",
    "StreamPartReasoningDelta",
    "StreamPartReasoningEnd",
    "StreamPartStreamStart",
    "StreamPartStreamEnd",
    "StreamPart",
    # tools & plugins
    "Plugin",
    "PluginInitError",
    "Skills",
    "Skill",
    "magic",
    "tool",
    "ImageUrl",
    "ProviderTool",
    "Tool",
    "Tools",
    "ToolResult",
    "MCP",
    "MCPContext",
    # utils
    "init_logging",
    "load_dotenv",
]



import abc
import re
from dataclasses import dataclass
from typing import AsyncGenerator, Sequence
import httpx
from agentia.llm import LLMOptions
from agentia.spec.chat import Message
from agentia.tools.tools import ToolSet
import agentia.spec as spec
from agentia.spec.stream import StreamPart


@dataclass
class GenerationResult:
    message: spec.AssistantMessage
    finish_reason: spec.FinishReason
    usage: spec.Usage
    provider_metadata: spec.ProviderMetadata | None


class Provider(abc.ABC):
    name: str
    model: str
    supported_urls: dict[str, Sequence[re.Pattern | str]]
    """
    Supported URL patterns by media type for the provider.

    The keys are media type patterns or full media types (e.g. `*\\/*` for everything, `audio/*`, `video/*`, or `application/pdf`).
    and the values are arrays of regular expressions that match the URL paths.

    The matching should be against lower-case URLs.

    Matched URLs are supported natively by the model and are not downloaded.
    """

    def __init__(self, name: str, model: str):
        self.name = name
        self.model = model
        self.supported_urls = {}

    @abc.abstractmethod
    async def generate(
        self,
        messages: list[Message],
        tools: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> GenerationResult: ...

    @abc.abstractmethod
    def stream(
        self,
        messages: list[Message],
        tools: ToolSet,
        options: LLMOptions,
        client: httpx.AsyncClient,
    ) -> AsyncGenerator[StreamPart, None]: ...



from . import Plugin, tool
import subprocess
from typing import Annotated, override
from pathlib import Path
import base64
import datetime
from ..tools import ToolResult
from ..spec.base import File


class Bash(Plugin):
    @override
    def get_instructions(self) -> str | None:
        s = f"CURRENT TIME: {datetime.datetime.now().isoformat()}\n"
        s += f"CURRENT WORKING DIRECTORY: `{str(Path.cwd())}`.\n"
        return s

    @tool(name="Bash")
    async def run_bash_command(
        self,
        command: Annotated[str, "The bash command to run"],
        cwd: Annotated[
            str | None, "The optional working directory for the command"
        ] = None,
    ) -> str:
        """Run a bash command and return the output."""
        await self.agent.user_consent_guard("Run command: " + command)
        result = subprocess.run(
            command, shell=True, capture_output=True, text=True, cwd=cwd
        )
        return result.stdout.strip()

    @tool(name="LoadFile")
    async def load_file(
        self, path: Annotated[str, "The path to the file to load."]
    ) -> ToolResult:
        """
        Load NON-TEXT files into the conversation so that you can inspect their content.
        Only images (.png, .jpg, .jpeg) and PDFs (.pdf) are supported. No directory is allowed, only file paths.
        For text files, directly read them using the `cat` command in Bash.
        """
        if not path:
            return ToolResult(output="No file path provided.")

        p = Path(path)

        if not p.exists():
            return ToolResult(output=f"File '{path}' does not exist.")

        if p.suffix.lower() in [".png", ".jpg", ".jpeg"]:
            media_type = f"image/png" if p.suffix.lower() == ".png" else f"image/jpeg"
        elif p.suffix.lower() == ".pdf":
            media_type = "application/pdf"
        else:
            return ToolResult(output=f"Unsupported file type: '{p.suffix}'.")

        bytes = p.read_bytes()
        base64_url = f"data:{media_type};base64,{base64.b64encode(bytes).decode()}"
        file = File(media_type=media_type, data=base64_url)
        return ToolResult(files=[file], output=f"Loaded file '{path}' successfully.")
